import contextlib
import dataclasses
import functools
import logging
import time
import urllib.parse
from collections.abc import Collection
from typing import Any

import aiohttp
import aiohttp_retry

from .base import sanitize_params
from .jsons import TJSONDumpable, TJSONParsed, json_loads

LOGGER = logging.getLogger(__name__)


def resp_data_for_debug(resp_data: bytes, max_length: int = 100, cut_suffix: str = "...") -> str:
    # Could avoid decoding resp_data beyond `max_length * 4` or so, but it shouldn't matter much.
    resp_text = resp_data.decode("utf-8", errors="replace")
    if len(resp_text) <= max_length:
        return repr(resp_text)
    return repr(f"{resp_text[:max_length]}{cut_suffix}")


@dataclasses.dataclass()
class HTTPResponse:
    """Simplified wrapper of `aiohttp.ClientResponse` with pre-read body"""

    orig: aiohttp.ClientResponse
    content: bytes
    time_taken_sec: float

    max_resp_log_len: int | None = None

    @property
    def status(self) -> int:
        return self.orig.status

    def json(self) -> TJSONParsed:
        return json_loads(self.content)

    def json_untyped(self) -> Any:
        return json_loads(self.content)

    @functools.cached_property
    def str_for_log(self) -> str:
        result = f"{self.status}, {len(self.content)}b in {self.time_taken_sec:.3f}s"
        if self.max_resp_log_len:
            debug_data = resp_data_for_debug(self.content, self.max_resp_log_len)
            result = f"{result} data={debug_data}"
        return result


@dataclasses.dataclass()
class HTTPClient:
    session: aiohttp.ClientSession | None = None
    _session_managed: bool = False
    logger: logging.Logger = LOGGER

    retry_statuses: Collection[int] = frozenset({429, *range(500, 600)})
    retry_attempts: int = 5
    retry_start_timeout_sec: float = 0.5
    default_timeout_sec: float = 5.0
    max_resp_log_len: int = 300
    max_resp_error_message_len: int = 1000

    @staticmethod
    def make_url_for_log(url: str, params: dict[str, Any]) -> str:
        """
        Join url with the query parameters, sanitizing the parameters as needed.

        Warning: Does not necessarily match the aiohttp's logic;
        should not be considered a securely correct representation.

        >>> make_url_for_log = HTTPClient.make_url_for_log
        >>> make_url_for_log("http://example.com", dict(val="a", apiKey="b" * 20))
        'http://example.com?val=a&apiKey=bb...bb'
        >>> make_url_for_log("http://example.com?val=0a&inparam=123", dict(val="a", apiKey="b" * 20))
        'http://example.com?val=0a&inparam=123&val=a&apiKey=bb...bb'
        """
        params_clean = sanitize_params(params)
        params_s = urllib.parse.urlencode(params_clean)
        return f"{url}&{params_s}" if "?" in url else f"{url}?{params_s}"

    @property
    def _retry_options(self) -> aiohttp_retry.RetryOptionsBase:
        return aiohttp_retry.ExponentialRetry(
            attempts=self.retry_attempts,
            start_timeout=self.retry_start_timeout_sec,
            max_timeout=self.default_timeout_sec,
            statuses=set(self.retry_statuses),
        )

    async def __aenter__(self) -> "HTTPClient":
        if self.session is None:
            assert not self._session_managed
            self._session_managed = True
            self.session = aiohttp.ClientSession()
            await self.session.__aenter__()

        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, exc_tb: Any) -> None:
        if self._session_managed:
            assert self.session is not None
            session = self.session
            self.session = None
            self._session_managed = False
            await session.__aexit__(exc_type, exc_value, exc_tb)

    async def close(self) -> None:
        await self.__aexit__(None, None, None)

    async def req(
        self,
        url: str,
        *,
        method: str = "get",
        params: dict[str, Any] | None = None,
        require_ok: bool = True,
        **kwargs: Any,
    ) -> HTTPResponse:
        do_logging = self.logger.isEnabledFor(logging.DEBUG)
        if do_logging:
            url_for_log = self.make_url_for_log(url, params or {})
            req_log_s = f"{method.upper()} {url_for_log}"
            self.logger.debug("HTTP request: %s", req_log_s)

        async with contextlib.AsyncExitStack() as acm:
            sess = self.session
            # Making the `__aenter__` of this client optional
            # (in which case it creates a session-per-request)
            if sess is None:
                sess = await acm.enter_async_context(aiohttp.ClientSession())

            retry_client = aiohttp_retry.RetryClient(sess, retry_options=self._retry_options)
            start_time = time.monotonic()
            async with retry_client.request(
                method, url, params=params or {}, timeout=self.default_timeout_sec, **kwargs
            ) as resp:
                resp_content = await resp.read()
                wrapped_resp = HTTPResponse(
                    orig=resp,
                    content=resp_content,
                    time_taken_sec=time.monotonic() - start_time,
                    max_resp_log_len=self.max_resp_log_len,
                )

                try:
                    if require_ok:
                        resp.raise_for_status()
                except aiohttp.ClientResponseError as exc:
                    resp_text_cut = resp_data_for_debug(resp_content, self.max_resp_error_message_len)
                    resp_text_msg = f"resp_text[:{self.max_resp_error_message_len}]={resp_text_cut}"
                    exc.message = f"{exc.message}; {resp_text_msg}"
                    if do_logging:
                        self.logger.debug(
                            "HTTP response exception: %s -> %s, %s", req_log_s, wrapped_resp.str_for_log, exc.message
                        )
                    raise

                if do_logging:
                    LOGGER.debug("HTTP response: %s -> %s", req_log_s, wrapped_resp.str_for_log)

        return wrapped_resp

    async def req_jsonrpc(
        self,
        url: str,
        jsonrpc_method: str,
        params: TJSONDumpable,
        request_id: str | None = None,
        jsonrpc_version: str | None = "2.0",
        **kwargs: Any,
    ) -> Any:
        data = {"method": jsonrpc_method, "params": params}
        if jsonrpc_version is not None:
            data["jsonrpc"] = jsonrpc_version
        if request_id is not None:
            data["id"] = request_id

        resp = await self.req(url, method="post", json=data, **kwargs)
        resp_data = resp.json_untyped()

        if not isinstance(resp_data, dict):
            raise ValueError(f"Unexpected JSONRPC response type: {type(resp_data)}, {repr(resp_data)[:1024]=!r}")
        # Not including any sanity validation ("jsonrpc", "id")
        result = resp_data.get("result")
        error = resp_data.get("error")

        if error is not None:
            if result is not None:
                LOGGER.warning("JSONRPC with both error and result: %r, %s", error, repr(result)[:2048])
                return result

            raise ValueError(f"JSONRPC error: {error!r}")

        return result



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, TypeVar

TKey = TypeVar("TKey")
TVal = TypeVar("TVal")


def ensure_unique_keys(items: Iterable[tuple[TKey, TVal]]) -> dict[TKey, TVal]:
    """Replacement for dict comprehension that checks the keys' uniqueness"""
    result: dict[TKey, TVal] = {}
    for key, value in items:
        if key in result:
            raise ValueError(f"Key conflict: {key=!r}, first_value={result[key]!r}, second_value={value!r}")
        result[key] = value
    return result


def pop_keys(data: dict[TKey, TVal], keys: Sequence[TKey]) -> dict[TKey, TVal]:
    result = data.copy()
    for key in keys:
        result.pop(key, None)
    return result


def clear_none(data: Mapping[TKey, TVal | None]) -> dict[TKey, TVal]:
    return {key: val for key, val in data.items() if val is not None}


def groupby(items: Iterable[tuple[TKey, TVal]]) -> dict[TKey, list[TVal]]:
    """
    Simple "group by" over iterable items.

    >>> groupby([(1, 1), (2, 2), (1, 3)])
    {1: [1, 3], 2: [2]}
    """
    res: dict[TKey, list[TVal]] = {}
    for key, val in items:
        try:
            group_list = res[key]
        except KeyError:
            res[key] = [val]
        else:
            group_list.append(val)
    return res


def _is_common_prefix(item_a: Sequence[Any], item_b: Sequence[Any]) -> bool:
    """
    Check whether one (either) element is a prefix of another.

    >>> _is_common_prefix("abc", "ab")
    True
    >>> _is_common_prefix("ab", "abc")
    True
    >>> _is_common_prefix("abc", "abd")
    False
    >>> _is_common_prefix("abc", "ax")
    False
    """
    min_len = min(len(item_a), len(item_b))
    return item_a[:min_len] == item_b[:min_len]


TGroupbyRecLeaf = list[TVal]
TGroupbyRecNode = dict[TKey, "TGroupbyRecNode | TGroupbyRecLeaf"]
TGroupbyRecRet = TGroupbyRecNode


def groupby_rec(items: Iterable[tuple[Sequence[TKey], TVal]]) -> TGroupbyRecRet:
    """
    Recursive "group by" over iterable items.

    >>> groupby_rec([([1, 1], 11), ([1, 2], 12), ([2, 1], 21)])
    {1: {1: [11], 2: [12]}, 2: {1: [21]}}
    >>> groupby_rec([([1, 1], 11), ([1, 2], 12), ([1], 2)])
    Traceback (most recent call last):
      ...
    ValueError: Key path (1,) conflicts with previously seen paths [(1, 1), (1, 2)]
    >>> groupby_rec([([1], 2), ([1, 1], 11)])
    Traceback (most recent call last):
      ...
    ValueError: Key path (1, 1) conflicts with previously seen paths [(1,)]
    """
    res: TGroupbyRecRet = {}
    seen: set[tuple[TKey, ...]] = set()
    for key_path_raw, val in items:
        key_path = tuple(key_path_raw)

        if not key_path:
            raise ValueError("Empty key path")
        conflicts = [seen_path for seen_path in seen if _is_common_prefix(seen_path, key_path)]
        if conflicts:
            raise ValueError(f"Key path {key_path!r} conflicts with previously seen paths {conflicts!r}")
        seen.add(key_path)

        here: TGroupbyRecNode = res
        for key in key_path[:-1]:
            if key in here:
                next_here = here[key]
                assert isinstance(next_here, dict)
                here = next_here
            else:
                new_here: TGroupbyRecNode = {}
                here[key] = new_here
                here = new_here

        leaf_key = key_path[-1]
        group_list: list[TVal]
        if leaf_key in here:
            here_leaf = here[leaf_key]
            assert isinstance(here_leaf, list)
            group_list = here_leaf
        else:
            group_list = []
            here[leaf_key] = group_list

        group_list.append(val)

    return res



import datetime


def dt_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.UTC)


def require_aware_dt(value: datetime.datetime) -> datetime.datetime:
    if value.tzinfo is None:
        raise ValueError(f"TZ-aware datetime is required, got {value!r}")
    return value


def floor_dt_to_interval(dt: datetime.datetime, interval_sec: float) -> datetime.datetime:
    """
    Round datetime down to the nearest whole number of intervals from midnight.

    >>> floor_dt_to_interval(datetime.datetime.fromisoformat("2023-04-05T06:07:48Z"), 30.0).isoformat()
    '2023-04-05T06:07:30+00:00'
    >>> floor_dt_to_interval(datetime.datetime.fromisoformat("2023-04-05T06:07:48Z"), 60.0).isoformat()
    '2023-04-05T06:07:00+00:00'
    """
    dt_date = dt.replace(hour=0, minute=0, second=0, microsecond=0)
    day_seconds = (dt - dt_date).total_seconds()
    assert day_seconds >= 0
    dt_intervals = day_seconds / interval_sec
    offset_sec = (dt_intervals - int(dt_intervals)) * interval_sec
    return dt - datetime.timedelta(seconds=offset_sec)



import dataclasses
import logging
import re
from contextvars import ContextVar

import starlette.datastructures
import starlette.types

from .trace import TRACE_ID_VAR, new_trace_id

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass()
class TraceIdMiddleware:
    app: starlette.types.ASGIApp
    header_name: str = "X-Request-ID"
    header_validation_re: str = r"^[a-z0-9]{12,20}$"
    trace_id_var: ContextVar[str | None] = TRACE_ID_VAR
    trace_id_prefix: str = "aa"  # "API App"

    def _new_trace_id(self, parent: str | None = None) -> str:
        return new_trace_id(self.trace_id_prefix, parent=parent)

    async def __call__(
        self, scope: starlette.types.Scope, receive: starlette.types.Receive, send: starlette.types.Send
    ) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        headers = starlette.datastructures.MutableHeaders(scope=scope)
        header_value = headers.get(self.header_name.lower())
        parent_trace_id = None
        if header_value:
            if re.search(self.header_validation_re, header_value):
                parent_trace_id = header_value
            else:
                LOGGER.debug("Invalid trace id in header %r: %r", self.header_name, header_value)

        trace_id = self._new_trace_id(parent=parent_trace_id)

        self.trace_id_var.set(trace_id)

        async def handle_outgoing_request(message: starlette.types.Message) -> None:
            local_trace_id = self.trace_id_var.get()
            if message["type"] == "http.response.start" and local_trace_id:
                headers = starlette.datastructures.MutableHeaders(scope=message)
                headers.append(self.header_name, local_trace_id)

            await send(message)

        await self.app(scope, receive, handle_outgoing_request)



from collections.abc import Callable
from typing import Annotated, Any, ClassVar, TypeVar

from frozendict import frozendict
from frozendict.cool import deepfreeze
from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core.core_schema import (
    CoreSchema,
    chain_schema,
    dict_schema,
    is_instance_schema,
    json_or_python_schema,
    no_info_plain_validator_function,
    plain_serializer_function_ser_schema,
    union_schema,
)


class _PydanticFrozenDictAnnotation:
    """
    https://github.com/pydantic/pydantic/discussions/8721
    """

    _cast_func: ClassVar[Callable[[Any], Any]] = frozendict

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, _handler: GetCoreSchemaHandler) -> CoreSchema:
        from_dict_schema = chain_schema([dict_schema(), no_info_plain_validator_function(cls._cast_func)])
        return json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=union_schema([is_instance_schema(frozendict), from_dict_schema]),
            serialization=plain_serializer_function_ser_schema(dict),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _: CoreSchema, handler: GetJsonSchemaHandler) -> JsonSchemaValue:
        return handler(dict_schema())


class _PydanticDeepFrozenDictAnnotation(_PydanticFrozenDictAnnotation):
    """
    A hack to handle multi-layered frozen dicts.
    Ideally, recursive use of the normal form should work,
    but no such luck.
    WARNING: does not support pydantic models inside the structure (forces them into frozendicts).
    """

    _cast_func: ClassVar[Callable[[Any], Any]] = deepfreeze


_K = TypeVar("_K")
_V = TypeVar("_V")
PydanticFrozenDict = Annotated[frozendict[_K, _V], _PydanticFrozenDictAnnotation]
PydanticDeepFrozenDict = Annotated[frozendict[_K, _V], _PydanticDeepFrozenDictAnnotation]



from typing import Any

import pydantic
import pydantic_settings
import yaml


class YAMLSettingsMixin(pydantic_settings.PydanticBaseSettingsSource):
    """Addition to allow YAML instead of JSON inside variables"""

    def decode_complex_value(self, field_name: str, field: pydantic.fields.FieldInfo, value: Any) -> Any:
        return yaml.safe_load(value)


class YAMLedEnvSettingsSource(YAMLSettingsMixin, pydantic_settings.EnvSettingsSource):
    """EnvSettings + YAML"""


class YAMLedDotEnvSettingsSource(YAMLSettingsMixin, pydantic_settings.DotEnvSettingsSource):
    """DotEnvSettings + YAML"""



import asyncio
import contextlib
import dataclasses
from asyncio.futures import Future
from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Iterable
from typing import Any, Generic, Literal, TypeAlias, TypeVar, cast, overload

from .trace import TRACE_ID_VAR

TRet = TypeVar("TRet")
TVal = TypeVar("TVal")


async def adummy(value: TVal) -> TVal:
    return value


async def named_await(name: str, func: Callable[[], Awaitable[TRet]]) -> tuple[str, TRet]:
    """Convenience helper for `asyncio.gather`ing into a dict of results"""
    result = await func()
    return (name, result)


def _get_task_child_next_id(attr_name: str = "_trace_child_next_id") -> int:
    task = asyncio.current_task()
    value = getattr(task, attr_name, None) or 0
    value += 1
    setattr(task, attr_name, value)
    return value


# Due to wrapped `asyncio.gather`,
# adapted from
# https://github.com/python/mypy/blob/ea49e1fa488810997d192a36d85357dadb4a7f14/mypy/typeshed/stdlib/asyncio/tasks.pyi#L99
_T = TypeVar("_T")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_T6 = TypeVar("_T6")
# For wrapping simplicity, omitting the `Generator[Any, None, _T]` from `_FutureLike`
_FutureLike: TypeAlias = Future[_T] | Awaitable[_T]


def _wrap_with_trace_id(coro: _FutureLike[TRet]) -> _FutureLike[TRet]:
    current_trace_id = TRACE_ID_VAR.get()
    if not current_trace_id:  # nothing to handle
        return coro

    child_id = _get_task_child_next_id()
    new_trace_id = f"{current_trace_id}__{child_id}"

    async def wrapped_with_trace_id() -> TRet:
        TRACE_ID_VAR.set(new_trace_id)
        return await coro

    return wrapped_with_trace_id()


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1], *, return_exceptions: Literal[False] = False
) -> tuple[_T1]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    __coro_or_future5: _FutureLike[_T5],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4, _T5]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    __coro_or_future5: _FutureLike[_T5],
    __coro_or_future6: _FutureLike[_T6],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4, _T5, _T6]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    *coros_or_futures: _FutureLike[_T], return_exceptions: Literal[False] = False
) -> list[_T]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1], *, return_exceptions: bool
) -> tuple[_T1 | BaseException]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1], __coro_or_future2: _FutureLike[_T2], *, return_exceptions: bool
) -> tuple[_T1 | BaseException, _T2 | BaseException]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    *,
    return_exceptions: bool,
) -> tuple[_T1 | BaseException, _T2 | BaseException, _T3 | BaseException]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    *,
    return_exceptions: bool,
) -> tuple[_T1 | BaseException, _T2 | BaseException, _T3 | BaseException, _T4 | BaseException]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    __coro_or_future5: _FutureLike[_T5],
    *,
    return_exceptions: bool,
) -> tuple[_T1 | BaseException, _T2 | BaseException, _T3 | BaseException, _T4 | BaseException, _T5 | BaseException]: ...


@overload
async def aiogather(  # type: ignore[overload-overlap]
    __coro_or_future1: _FutureLike[_T1],
    __coro_or_future2: _FutureLike[_T2],
    __coro_or_future3: _FutureLike[_T3],
    __coro_or_future4: _FutureLike[_T4],
    __coro_or_future5: _FutureLike[_T5],
    __coro_or_future6: _FutureLike[_T6],
    *,
    return_exceptions: bool,
) -> tuple[
    _T1 | BaseException,
    _T2 | BaseException,
    _T3 | BaseException,
    _T4 | BaseException,
    _T5 | BaseException,
    _T6 | BaseException,
]: ...


@overload
async def aiogather(*coros_or_futures: _FutureLike[_T], return_exceptions: bool) -> list[_T | BaseException]: ...


async def aiogather(
    *coros_or_futures: _FutureLike[_T], return_exceptions: bool = False
) -> list[_T] | list[BaseException] | list[_T | BaseException] | tuple[_T | BaseException, ...]:
    """
    Wrapper around `asyncio.gather` (wihout extra parameters) with various conveniences.

    NOTE: takes a single argument, rather than a coro-per-arg.
    This makes gathering from a list/generator comprehension easier.
    """
    if not coros_or_futures:
        return []

    if len(coros_or_futures) == 1:
        try:
            result = await coros_or_futures[0]
        except BaseException as exc:
            if return_exceptions:
                return [exc]
            raise
        return [result]

    return await asyncio.gather(*[_wrap_with_trace_id(coro) for coro in coros_or_futures])


def aiogather_it(coros_or_futures: Iterable[_FutureLike[_T]]) -> Awaitable[list[_T]]:
    return aiogather(*coros_or_futures)


@contextlib.asynccontextmanager
async def task_cm(coro_or_future: Coroutine | Future) -> AsyncGenerator[asyncio.Task, None]:
    task = asyncio.ensure_future(coro_or_future)
    try:
        yield task
    finally:
        if not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


TArgsKey = tuple[
    # args
    tuple[Any, ...],
    # kwargs
    tuple[tuple[str, Any], ...],
]
TAIOMemoStorage = dict[TArgsKey, tuple[asyncio.Lock, None] | tuple[None, TRet] | None]


@dataclasses.dataclass()
class AIOLockedMemoized(Generic[TRet]):
    """
    Storage states:

        none: `None`
        none -> locked
        locked: `Lock, None`
        locked -> saved
        locked -> error
        saved: `None, TRet`
        error: `None`
        error -> none
    """

    func: Callable[..., Awaitable[TRet]]
    miss_count: int = 0
    wait_count: int = 0
    hit_count: int = 0
    error_in_count: int = 0
    error_count: int = 0

    def __post_init__(self) -> None:
        func_name = self.func.__name__
        if "<" in func_name:
            raise ValueError(f"Attempting to memoize a non-normal function with name={func_name!r}")
        self._storage_key = f"_{func_name}_cache"

    @staticmethod
    def _get_args_key(args: tuple[Any, ...], kwargs: dict[str, Any]) -> TArgsKey:
        return args, tuple(sorted(kwargs.items()))

    def _get_storage(self, obj: Any) -> TAIOMemoStorage[TRet]:
        data: TAIOMemoStorage[TRet]
        storage = getattr(obj, self._storage_key, None)
        if storage is not None:
            assert isinstance(storage, tuple)
            assert len(storage) == 2
            owner, data = storage
            assert owner is self
            assert data is not None
            return data

        data = {}
        setattr(obj, self._storage_key, (self, data))
        return data

    async def __call__(self, obj: Any, *args: Any, **kwargs: Any) -> TRet:
        args_key = self._get_args_key(args, kwargs)
        storage = self._get_storage(obj)

        state = storage.get(args_key)
        lock: asyncio.Lock | None

        if state is None:  # `none`
            self.miss_count += 1

        else:  # `locked` | `saved`
            assert isinstance(state, tuple)
            assert len(state) == 2
            lock, pre_result = state

            if lock is None:  # `saved`
                self.hit_count += 1
                # It can be `result: TRet = None`; for correct typing, might have to use a sentinel object.
                return cast(TRet, pre_result)

            # `locked`
            assert pre_result is None
            self.wait_count += 1
            async with lock:
                state = storage[args_key]

            if state is not None:  # `saved` | `error`->`locked`
                assert isinstance(state, tuple)
                assert len(state) == 2
                lock, pre_result = state
                if lock is None:
                    return cast(TRet, pre_result)

            # `error`
            self.error_count += 1

            # NOTE: in case of error, all lock-waiters will go generate their own locks in parallel.

        # `none` | `error` | `error`->`locked`
        lock = asyncio.Lock()
        state = (lock, None)
        if storage.get(args_key) is None:
            storage[args_key] = state
        async with lock:
            try:
                result = await self.func(obj, *args, **kwargs)
            except Exception:
                self.error_in_count += 1
                # Another process could have generated a valid state,
                # if the previous state was `error`.
                # Thus, don't clean the state that wasn't our own.
                if storage.get(args_key) is state:
                    storage[args_key] = None
                raise

            storage[args_key] = (None, result)

        return result


def aio_locked_memoized(
    *,
    memo_cls: type[AIOLockedMemoized] = AIOLockedMemoized,
) -> Callable[[Callable[..., Awaitable[TRet]]], Callable[..., Awaitable[TRet]]]:
    def aio_locked_memoized_wrap(func: Callable[..., Awaitable[TRet]]) -> Callable[..., Awaitable[TRet]]:
        memo = memo_cls(func=func)

        async def aio_locked_memoized_wrapped(self: Any, *args: Any, **kwargs: Any) -> TRet:
            return await memo(self, *args, **kwargs)

        setattr(aio_locked_memoized_wrapped, "memo", memo)
        return aio_locked_memoized_wrapped

    return aio_locked_memoized_wrap



import asyncio
import enum
import logging
import socket
from typing import Any, ClassVar

import yaml

from .trace import TRACE_ID_VAR

HOSTNAME = socket.gethostname()


# http://docs.python.org/library/logging.html#logrecord-attributes
LOGRECORD_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
}


class NotAvailable(enum.Enum):
    NOT_AVAILABLE = "NOT_AVAILABLE"


NOT_AVAILABLE = NotAvailable.NOT_AVAILABLE


class Annotator(logging.Filter):
    """A convenience abstract class for most annotators"""

    default_attribute_name: ClassVar[str | None] = None
    attribute_name: str
    force_overwrite: bool = False

    def __init__(self, *args: Any, attribute_name: str | None = None, **kwargs: Any) -> None:
        if not attribute_name:
            if not self.default_attribute_name:
                raise TypeError("attribute_name should either be on class or always specified")
            attribute_name = self.default_attribute_name
        self.attribute_name = attribute_name
        super().__init__(*args, **kwargs)

    def get_value(self, record: logging.LogRecord) -> Any:
        raise NotImplementedError

    def filter(self, record: logging.LogRecord) -> bool:
        """“annotate”, actually"""
        if hasattr(record, self.attribute_name) and not self.force_overwrite:
            return True
        value = self.get_value(record)
        setattr(record, self.attribute_name, value)
        return True


class HostnameAnnotator(Annotator):
    default_attribute_name: ClassVar[str] = "hostname"

    def get_value(self, record: logging.LogRecord) -> str:
        return HOSTNAME


class AsyncioTaskAnnotator(Annotator):
    default_attribute_name: ClassVar[str] = "aio_task"

    def get_value(self, record: logging.LogRecord) -> str | None:
        try:
            task = asyncio.current_task()
            if task:
                return task.get_name()
        except RuntimeError:
            pass

        return None


class TraceVarAnnotator(Annotator):
    default_attribute_name: ClassVar[str] = "trace_id"

    def get_value(self, record: logging.LogRecord) -> str | None:
        return TRACE_ID_VAR.get()


class ExtrasAnnotator(Annotator):
    default_attribute_name: ClassVar[str] = "_extras"
    width: int = 100

    def serialize_data(self, data: dict[str, Any]) -> str:
        return yaml.dump(
            data,
            Dumper=getattr(yaml, "CDumper", None) or yaml.Dumper,
            default_flow_style=None,
            sort_keys=False,
            width=self.width,
        )

    @staticmethod
    def postprocess_serialized_data(data: str) -> str:
        data = data.strip()
        if not data:
            return ""
        if "\n" not in data:
            return f"    {data}"
        indent = " " * 4
        return f"\n{indent}" + data.strip().replace("\n", f"\n{indent}")

    def get_value(self, record: logging.LogRecord) -> str:
        data: dict[str, Any] = {
            key: value
            for key, value in record.__dict__.items()
            if key not in LOGRECORD_ATTRS and not (hasattr(key, "startswith") and key.startswith("_"))
        }
        if not data:
            return ""
        data_s = self.serialize_data(data)
        return self.postprocess_serialized_data(data_s)
